"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ImportMap = void 0;
const DEFAULT_MODULE_MAP = {
    // External.
    solanaAccounts: '@solana/accounts',
    solanaAddresses: '@solana/addresses',
    solanaCodecsCore: '@solana/codecs',
    solanaCodecsNumbers: '@solana/codecs',
    solanaCodecsStrings: '@solana/codecs',
    solanaCodecsDataStructures: '@solana/codecs',
    solanaInstructions: '@solana/instructions',
    solanaPrograms: '@solana/programs',
    solanaOptions: '@solana/codecs',
    solanaSigners: '@solana/signers',
    // Internal.
    types: '../types',
    errors: '../errors',
    shared: '../shared',
    generated: '..',
    hooked: '../../hooked',
    generatedPrograms: '../programs',
    generatedPdas: '../pdas',
    generatedAccounts: '../accounts',
    generatedErrors: '../errors',
    generatedTypes: '../types',
    generatedInstructions: '../instructions',
};
class ImportMap {
    constructor() {
        this._imports = new Map();
        this._aliases = new Map();
    }
    add(module, imports) {
        var _a;
        const newImports = new Set(typeof imports === 'string' ? [imports] : imports);
        if (newImports.size === 0)
            return this;
        const currentImports = (_a = this._imports.get(module)) !== null && _a !== void 0 ? _a : new Set();
        newImports.forEach((i) => currentImports.add(i));
        this._imports.set(module, currentImports);
        return this;
    }
    remove(module, imports) {
        var _a;
        const importsToRemove = new Set(typeof imports === 'string' ? [imports] : imports);
        if (importsToRemove.size === 0)
            return this;
        const currentImports = (_a = this._imports.get(module)) !== null && _a !== void 0 ? _a : new Set();
        importsToRemove.forEach((i) => currentImports.delete(i));
        if (currentImports.size === 0) {
            this._imports.delete(module);
        }
        else {
            this._imports.set(module, currentImports);
        }
        return this;
    }
    mergeWith(...others) {
        others.forEach((rawOther) => {
            const other = 'imports' in rawOther ? rawOther.imports : rawOther;
            other._imports.forEach((imports, module) => {
                this.add(module, imports);
            });
            other._aliases.forEach((aliases, module) => {
                Object.entries(aliases).forEach(([name, alias]) => {
                    this.addAlias(module, name, alias);
                });
            });
        });
        return this;
    }
    mergeWithManifest(manifest) {
        return this.mergeWith(manifest.strictType, manifest.looseType, manifest.encoder, manifest.decoder);
    }
    addAlias(module, name, alias) {
        var _a;
        const currentAliases = (_a = this._aliases.get(module)) !== null && _a !== void 0 ? _a : {};
        currentAliases[name] = alias;
        this._aliases.set(module, currentAliases);
        return this;
    }
    isEmpty() {
        return this._imports.size === 0;
    }
    resolve(dependencies = {}) {
        // Resolve aliases.
        const aliasedMap = new Map([...this._imports.entries()].map(([module, imports]) => {
            var _a;
            const aliasMap = (_a = this._aliases.get(module)) !== null && _a !== void 0 ? _a : {};
            const joinedImports = [...imports].map((i) => aliasMap[i] ? `${i} as ${aliasMap[i]}` : i);
            return [module, new Set(joinedImports)];
        }));
        // Resolve dependency mappings.
        const dependencyMap = { ...DEFAULT_MODULE_MAP, ...dependencies };
        const resolvedMap = new Map();
        aliasedMap.forEach((imports, module) => {
            var _a, _b;
            const resolvedModule = (_a = dependencyMap[module]) !== null && _a !== void 0 ? _a : module;
            const currentImports = (_b = resolvedMap.get(resolvedModule)) !== null && _b !== void 0 ? _b : new Set();
            imports.forEach((i) => currentImports.add(i));
            resolvedMap.set(resolvedModule, currentImports);
        });
        return resolvedMap;
    }
    toString(dependencies = {}) {
        return [...this.resolve(dependencies).entries()]
            .sort(([a], [b]) => {
            const aIsRelative = a.startsWith('.');
            const bIsRelative = b.startsWith('.');
            if (aIsRelative && !bIsRelative)
                return 1;
            if (!aIsRelative && bIsRelative)
                return -1;
            return a.localeCompare(b);
        })
            .map(([module, imports]) => {
            const joinedImports = [...imports].sort().join(', ');
            return `import { ${joinedImports} } from '${module}';`;
        })
            .join('\n');
    }
}
exports.ImportMap = ImportMap;
//# sourceMappingURL=ImportMap.js.map