"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDiscriminatorConditionFragment = void 0;
const codecs_strings_1 = require("@solana/codecs-strings");
const nodes_1 = require("../../../nodes");
const shared_1 = require("../../../shared");
const visitors_1 = require("../../../visitors");
const common_1 = require("./common");
/**
 * ```
 * if (data.length === 72) {
 *   return splTokenAccounts.TOKEN;
 * }
 *
 * if (memcmp(data, getU32Encoder().encode(42), offset)) {
 *   return splTokenAccounts.TOKEN;
 * }
 *
 * if (memcmp(data, new Uint8Array([1, 2, 3]), offset)) {
 *   return splTokenAccounts.TOKEN;
 * }
 * ```
 */
function getDiscriminatorConditionFragment(scope) {
    return (0, common_1.mergeFragments)(scope.discriminators.flatMap((discriminator) => {
        if ((0, nodes_1.isNode)(discriminator, 'sizeDiscriminatorNode')) {
            return [getSizeConditionFragment(discriminator, scope)];
        }
        if ((0, nodes_1.isNode)(discriminator, 'constantDiscriminatorNode')) {
            return [getByteConditionFragment(discriminator, scope)];
        }
        if ((0, nodes_1.isNode)(discriminator, 'fieldDiscriminatorNode')) {
            return [getFieldConditionFragment(discriminator, scope)];
        }
        return [];
    }), (r) => r.join(' && ')).mapRender((r) => `if (${r}) { ${scope.ifTrue}; }`);
}
exports.getDiscriminatorConditionFragment = getDiscriminatorConditionFragment;
function getSizeConditionFragment(discriminator, scope) {
    const { dataName } = scope;
    return (0, common_1.fragment)(`${dataName}.length === ${discriminator.size}`);
}
function getByteConditionFragment(discriminator, scope) {
    const { dataName, typeManifestVisitor } = scope;
    const constant = (0, visitors_1.visit)(discriminator.constant, typeManifestVisitor).value;
    return constant
        .mapRender((r) => `memcmp(${dataName}, ${r}, ${discriminator.offset})`)
        .addImports('shared', 'memcmp');
}
function getFieldConditionFragment(discriminator, scope) {
    const field = scope.struct.fields.find((f) => f.name === discriminator.name);
    if (!field || !field.defaultValue) {
        throw new shared_1.InvalidKinobiTreeError(`Field discriminator "${discriminator.name}" does not have a matching argument with default value.`);
    }
    // This handles the case where a field uses an u8 array to represent its discriminator.
    // In this case, we can simplify the generated code by delegating to a constantDiscriminatorNode.
    if ((0, nodes_1.isNode)(field.type, 'arrayTypeNode') &&
        (0, nodes_1.isNode)(field.type.item, 'numberTypeNode') &&
        field.type.item.format === 'u8' &&
        (0, nodes_1.isNode)(field.type.count, 'fixedCountNode') &&
        (0, nodes_1.isNode)(field.defaultValue, 'arrayValueNode') &&
        field.defaultValue.items.every((0, nodes_1.isNodeFilter)('numberValueNode'))) {
        const base64Bytes = (0, codecs_strings_1.getBase64Decoder)().decode(new Uint8Array(field.defaultValue.items.map((node) => node.number)));
        return getByteConditionFragment((0, nodes_1.constantDiscriminatorNode)((0, nodes_1.constantValueNodeFromBytes)('base64', base64Bytes), discriminator.offset), scope);
    }
    return getByteConditionFragment((0, nodes_1.constantDiscriminatorNode)((0, nodes_1.constantValueNode)(field.type, field.defaultValue), discriminator.offset), scope);
}
//# sourceMappingURL=discriminatorCondition.js.map