"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionRemainingAccountsFragment = void 0;
const shared_1 = require("../../../shared");
const nodes_1 = require("../../../nodes");
const common_1 = require("./common");
function getInstructionRemainingAccountsFragment(scope) {
    const { remainingAccounts } = scope.instructionNode;
    const fragments = (remainingAccounts !== null && remainingAccounts !== void 0 ? remainingAccounts : []).flatMap((r) => getRemainingAccountsFragment(r, scope));
    if (fragments.length === 0)
        return (0, common_1.fragment)('');
    return (0, common_1.mergeFragments)(fragments, (r) => `// Remaining accounts.\n` +
        `const remainingAccounts: IAccountMeta[] = ${r.length === 1 ? r[0] : `[...${r.join(', ...')}]`}`).addImports('solanaInstructions', ['IAccountMeta']);
}
exports.getInstructionRemainingAccountsFragment = getInstructionRemainingAccountsFragment;
function getRemainingAccountsFragment(remainingAccounts, scope) {
    const remainingAccountsFragment = (() => {
        if ((0, nodes_1.isNode)(remainingAccounts.value, 'argumentValueNode')) {
            return getArgumentValueNodeFragment(remainingAccounts, scope);
        }
        if ((0, nodes_1.isNode)(remainingAccounts.value, 'resolverValueNode')) {
            return getResolverValueNodeFragment(remainingAccounts, scope);
        }
        return null;
    })();
    if (remainingAccountsFragment === null)
        return [];
    return [remainingAccountsFragment];
}
function getArgumentValueNodeFragment(remainingAccounts, scope) {
    var _a, _b, _c;
    const { instructionNode } = scope;
    (0, nodes_1.assertIsNode)(remainingAccounts.value, 'argumentValueNode');
    const argumentName = (0, shared_1.camelCase)(remainingAccounts.value.name);
    const isOptional = (_a = remainingAccounts.isOptional) !== null && _a !== void 0 ? _a : false;
    const isSigner = (_b = remainingAccounts.isSigner) !== null && _b !== void 0 ? _b : false;
    const isWritable = (_c = remainingAccounts.isWritable) !== null && _c !== void 0 ? _c : false;
    const nonSignerRole = isWritable
        ? 'AccountRole.WRITABLE'
        : 'AccountRole.READONLY';
    const signerRole = isWritable
        ? 'AccountRole.WRITABLE_SIGNER'
        : 'AccountRole.READONLY_SIGNER';
    const role = isSigner === true ? signerRole : nonSignerRole;
    const argumentArray = isOptional
        ? `(args.${argumentName} ?? [])`
        : `args.${argumentName}`;
    // The argument already exists or was added as `Array<Address>`.
    const allArguments = (0, nodes_1.getAllInstructionArguments)(instructionNode);
    const argumentExists = allArguments.some((arg) => arg.name === remainingAccounts.value.name);
    if (argumentExists || isSigner === false) {
        return (0, common_1.fragment)(`${argumentArray}.map((address) => ({ address, role: ${role} }))`).addImports('solanaInstructions', ['AccountRole']);
    }
    // The argument was added as `Array<TransactionSigner | Address>`.
    if (isSigner === 'either') {
        return (0, common_1.fragment)(`${argumentArray}.map((addressOrSigner) => (` +
            `isTransactionSigner(addressOrSigner)\n` +
            `? { address: addressOrSigner.address, role: ${role}, signer: addressOrSigner }\n` +
            `: { address: addressOrSigner, role: ${role} }\n` +
            `))`)
            .addImports('solanaInstructions', ['AccountRole'])
            .addImports('shared', ['isTransactionSigner']);
    }
    // The argument was added as `Array<TransactionSigner>`.
    return (0, common_1.fragment)(`${argumentArray}.map((signer) => ({ address: signer.address, role: ${signerRole}, signer }))`).addImports('solanaInstructions', ['AccountRole']);
}
function getResolverValueNodeFragment(remainingAccounts, scope) {
    var _a;
    (0, nodes_1.assertIsNode)(remainingAccounts.value, 'resolverValueNode');
    const isAsync = scope.asyncResolvers.includes(remainingAccounts.value.name);
    if (!scope.useAsync && isAsync)
        return null;
    const awaitKeyword = scope.useAsync && isAsync ? 'await ' : '';
    const functionName = scope.nameApi.resolverFunction(remainingAccounts.value.name);
    return (0, common_1.fragment)(`${awaitKeyword}${functionName}(resolverScope)`)
        .addImports((_a = remainingAccounts.value.importFrom) !== null && _a !== void 0 ? _a : 'hooked', functionName)
        .addFeatures(['instruction:resolverScopeVariable']);
}
//# sourceMappingURL=instructionRemainingAccounts.js.map