"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRenderMapVisitor = void 0;
const sync_1 = require("@prettier/sync");
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const customDataHelpers_1 = require("./customDataHelpers");
const fragments_1 = require("./fragments");
const getTypeManifestVisitor_1 = require("./getTypeManifestVisitor");
const ImportMap_1 = require("./ImportMap");
const nameTransformers_1 = require("./nameTransformers");
const DEFAULT_PRETTIER_OPTIONS = {
    semi: true,
    singleQuote: true,
    trailingComma: 'es5',
    useTabs: false,
    tabWidth: 2,
    arrowParens: 'always',
    printWidth: 80,
    parser: 'typescript',
};
function getRenderMapVisitor(options = {}) {
    var _a, _b, _c, _d, _e, _f, _g, _h;
    const linkables = new shared_1.LinkableDictionary();
    let program = null;
    const nameTransformers = {
        ...nameTransformers_1.DEFAULT_NAME_TRANSFORMERS,
        ...options.nameTransformers,
    };
    const nameApi = (0, nameTransformers_1.getNameApi)(nameTransformers);
    const renderParentInstructions = (_a = options.renderParentInstructions) !== null && _a !== void 0 ? _a : false;
    const formatCode = (_b = options.formatCode) !== null && _b !== void 0 ? _b : true;
    const prettierOptions = {
        ...DEFAULT_PRETTIER_OPTIONS,
        ...options.prettierOptions,
    };
    const dependencyMap = (_c = options.dependencyMap) !== null && _c !== void 0 ? _c : {};
    const asyncResolvers = ((_d = options.asyncResolvers) !== null && _d !== void 0 ? _d : []).map(shared_1.mainCase);
    const nonScalarEnums = ((_e = options.nonScalarEnums) !== null && _e !== void 0 ? _e : []).map(shared_1.mainCase);
    const internalNodes = ((_f = options.internalNodes) !== null && _f !== void 0 ? _f : []).map(shared_1.mainCase);
    const customAccountData = (0, customDataHelpers_1.parseCustomDataOptions)((_g = options.customAccountData) !== null && _g !== void 0 ? _g : [], 'AccountData');
    const customInstructionData = (0, customDataHelpers_1.parseCustomDataOptions)((_h = options.customInstructionData) !== null && _h !== void 0 ? _h : [], 'InstructionData');
    const getTypeManifestVisitor = (parentName) => (0, getTypeManifestVisitor_1.getTypeManifestVisitor)({
        nameApi,
        linkables,
        nonScalarEnums,
        customAccountData,
        customInstructionData,
        parentName,
    });
    const typeManifestVisitor = getTypeManifestVisitor();
    const resolvedInstructionInputVisitor = (0, visitors_1.getResolvedInstructionInputsVisitor)();
    const globalScope = {
        nameApi,
        linkables,
        typeManifestVisitor,
        asyncResolvers,
        nonScalarEnums,
        renderParentInstructions,
        customAccountData,
        customInstructionData,
    };
    const render = (template, context, renderOptions) => {
        const code = (0, shared_1.resolveTemplate)(`${__dirname}/templates`, template, context, renderOptions);
        return formatCode ? (0, sync_1.format)(code, prettierOptions) : code;
    };
    return (0, shared_1.pipe)((0, visitors_1.staticVisitor)(() => new shared_1.RenderMap(), [
        'rootNode',
        'programNode',
        'pdaNode',
        'accountNode',
        'definedTypeNode',
        'instructionNode',
    ]), (v) => (0, visitors_1.extendVisitor)(v, {
        visitRoot(node, { self }) {
            const isNotInternal = (n) => !internalNodes.includes(n.name);
            const programsToExport = (0, nodes_1.getAllPrograms)(node).filter(isNotInternal);
            const programsWithErrorsToExport = programsToExport.filter((p) => p.errors.length > 0);
            const pdasToExport = (0, nodes_1.getAllPdas)(node);
            const accountsToExport = (0, nodes_1.getAllAccounts)(node).filter(isNotInternal);
            const instructionsToExport = (0, nodes_1.getAllInstructionsWithSubs)(node, {
                leavesOnly: !renderParentInstructions,
            }).filter(isNotInternal);
            const definedTypesToExport = (0, nodes_1.getAllDefinedTypes)(node).filter(isNotInternal);
            const hasAnythingToExport = programsToExport.length > 0 ||
                accountsToExport.length > 0 ||
                instructionsToExport.length > 0 ||
                definedTypesToExport.length > 0;
            const ctx = {
                root: node,
                programsToExport,
                programsWithErrorsToExport,
                pdasToExport,
                accountsToExport,
                instructionsToExport,
                definedTypesToExport,
                hasAnythingToExport,
            };
            const map = new shared_1.RenderMap();
            if (hasAnythingToExport) {
                map.add('shared/index.ts', render('sharedPage.njk', ctx));
            }
            if (programsToExport.length > 0) {
                map.add('programs/index.ts', render('programsIndex.njk', ctx));
            }
            if (programsWithErrorsToExport.length > 0) {
                map.add('errors/index.ts', render('errorsIndex.njk', ctx));
            }
            if (accountsToExport.length > 0) {
                map.add('accounts/index.ts', render('accountsIndex.njk', ctx));
            }
            if (pdasToExport.length > 0) {
                map.add('pdas/index.ts', render('pdasIndex.njk', ctx));
            }
            if (instructionsToExport.length > 0) {
                map.add('instructions/index.ts', render('instructionsIndex.njk', ctx));
            }
            if (definedTypesToExport.length > 0) {
                map.add('types/index.ts', render('definedTypesIndex.njk', ctx));
            }
            return map
                .add('index.ts', render('rootIndex.njk', ctx))
                .add('global.d.ts', render('globalTypesPage.njk', ctx))
                .mergeWith(...(0, nodes_1.getAllPrograms)(node).map((p) => (0, visitors_1.visit)(p, self)));
        },
        visitProgram(node, { self }) {
            program = node;
            const customDataDefinedType = [
                ...(0, customDataHelpers_1.getDefinedTypeNodesToExtract)(node.accounts, customAccountData),
                ...(0, customDataHelpers_1.getDefinedTypeNodesToExtract)(node.instructions, customInstructionData),
            ];
            const scope = { ...globalScope, programNode: node };
            const renderMap = new shared_1.RenderMap()
                .mergeWith(...node.pdas.map((p) => (0, visitors_1.visit)(p, self)))
                .mergeWith(...node.accounts.map((a) => (0, visitors_1.visit)(a, self)))
                .mergeWith(...node.definedTypes.map((t) => (0, visitors_1.visit)(t, self)))
                .mergeWith(...customDataDefinedType.map((t) => (0, visitors_1.visit)(t, self)));
            if (node.errors.length > 0) {
                const programErrorsFragment = (0, fragments_1.getProgramErrorsFragment)(scope);
                renderMap.add(`errors/${(0, shared_1.camelCase)(node.name)}.ts`, render('errorsPage.njk', {
                    imports: new ImportMap_1.ImportMap()
                        .mergeWith(programErrorsFragment)
                        .toString(dependencyMap),
                    programErrorsFragment,
                }));
            }
            const programFragment = (0, fragments_1.getProgramFragment)(scope);
            const programAccountsFragment = (0, fragments_1.getProgramAccountsFragment)(scope);
            const programInstructionsFragment = (0, fragments_1.getProgramInstructionsFragment)(scope);
            renderMap.add(`programs/${(0, shared_1.camelCase)(node.name)}.ts`, render('programsPage.njk', {
                imports: new ImportMap_1.ImportMap()
                    .mergeWith(programFragment, programAccountsFragment, programInstructionsFragment)
                    .toString(dependencyMap),
                programFragment,
                programAccountsFragment,
                programInstructionsFragment,
            }));
            renderMap.mergeWith(...(0, nodes_1.getAllInstructionsWithSubs)(program, {
                leavesOnly: !renderParentInstructions,
            }).map((ix) => (0, visitors_1.visit)(ix, self)));
            program = null;
            return renderMap;
        },
        visitPda(node) {
            if (!program) {
                throw new Error('Account must be visited inside a program.');
            }
            const scope = { ...globalScope, pdaNode: node, programNode: program };
            const pdaFunctionFragment = (0, fragments_1.getPdaFunctionFragment)(scope);
            const imports = new ImportMap_1.ImportMap().mergeWith(pdaFunctionFragment);
            return new shared_1.RenderMap().add(`pdas/${(0, shared_1.camelCase)(node.name)}.ts`, render('pdasPage.njk', {
                imports: imports.toString(dependencyMap),
                pdaFunctionFragment,
            }));
        },
        visitAccount(node) {
            if (!program) {
                throw new Error('Account must be visited inside a program.');
            }
            const scope = {
                ...globalScope,
                accountNode: node,
                programNode: program,
                typeManifest: (0, visitors_1.visit)(node, typeManifestVisitor),
            };
            const accountTypeFragment = (0, fragments_1.getAccountTypeFragment)(scope);
            const accountFetchHelpersFragment = (0, fragments_1.getAccountFetchHelpersFragment)(scope);
            const accountSizeHelpersFragment = (0, fragments_1.getAccountSizeHelpersFragment)(scope);
            const accountPdaHelpersFragment = (0, fragments_1.getAccountPdaHelpersFragment)(scope);
            const imports = new ImportMap_1.ImportMap().mergeWith(accountTypeFragment, accountFetchHelpersFragment, accountSizeHelpersFragment, accountPdaHelpersFragment);
            return new shared_1.RenderMap().add(`accounts/${(0, shared_1.camelCase)(node.name)}.ts`, render('accountsPage.njk', {
                imports: imports.toString(dependencyMap),
                accountTypeFragment,
                accountFetchHelpersFragment,
                accountSizeHelpersFragment,
                accountPdaHelpersFragment,
            }));
        },
        visitInstruction(node) {
            var _a;
            if (!program) {
                throw new Error('Instruction must be visited inside a program.');
            }
            const instructionExtraName = nameApi.instructionExtraType(node.name);
            const scope = {
                ...globalScope,
                instructionNode: node,
                programNode: program,
                renamedArgs: getRenamedArgsMap(node),
                dataArgsManifest: (0, visitors_1.visit)(node, typeManifestVisitor),
                extraArgsManifest: (0, visitors_1.visit)((0, nodes_1.structTypeNodeFromInstructionArgumentNodes)((_a = node.extraArguments) !== null && _a !== void 0 ? _a : []), getTypeManifestVisitor({
                    strict: nameApi.dataType(instructionExtraName),
                    loose: nameApi.dataArgsType(instructionExtraName),
                })),
                resolvedInputs: (0, visitors_1.visit)(node, resolvedInstructionInputVisitor),
            };
            // Fragments.
            const instructionTypeFragment = (0, fragments_1.getInstructionTypeFragment)(scope);
            const instructionDataFragment = (0, fragments_1.getInstructionDataFragment)(scope);
            const instructionExtraArgsFragment = (0, fragments_1.getInstructionExtraArgsFragment)(scope);
            const instructionFunctionAsyncFragment = (0, fragments_1.getInstructionFunctionFragment)({
                ...scope,
                useAsync: true,
            });
            const instructionFunctionSyncFragment = (0, fragments_1.getInstructionFunctionFragment)({
                ...scope,
                useAsync: false,
            });
            const instructionParseFunctionFragment = (0, fragments_1.getInstructionParseFunctionFragment)(scope);
            // Imports and interfaces.
            const imports = new ImportMap_1.ImportMap().mergeWith(instructionTypeFragment, instructionDataFragment, instructionExtraArgsFragment, instructionFunctionAsyncFragment, instructionFunctionSyncFragment, instructionParseFunctionFragment);
            return new shared_1.RenderMap().add(`instructions/${(0, shared_1.camelCase)(node.name)}.ts`, render('instructionsPage.njk', {
                instruction: node,
                imports: imports.toString(dependencyMap),
                instructionTypeFragment,
                instructionDataFragment,
                instructionExtraArgsFragment,
                instructionFunctionAsyncFragment,
                instructionFunctionSyncFragment,
                instructionParseFunctionFragment,
            }));
        },
        visitDefinedType(node) {
            const scope = {
                ...globalScope,
                typeNode: node.type,
                name: node.name,
                manifest: (0, visitors_1.visit)(node, typeManifestVisitor),
                typeDocs: node.docs,
                encoderDocs: [],
                decoderDocs: [],
                codecDocs: [],
            };
            const typeWithCodecFragment = (0, fragments_1.getTypeWithCodecFragment)(scope);
            const typeDiscriminatedUnionHelpersFragment = (0, fragments_1.getTypeDiscriminatedUnionHelpersFragment)(scope);
            const imports = new ImportMap_1.ImportMap()
                .mergeWith(typeWithCodecFragment, typeDiscriminatedUnionHelpersFragment)
                .remove('generatedTypes', [
                nameApi.dataType(node.name),
                nameApi.dataArgsType(node.name),
                nameApi.encoderFunction(node.name),
                nameApi.decoderFunction(node.name),
                nameApi.codecFunction(node.name),
            ]);
            return new shared_1.RenderMap().add(`types/${(0, shared_1.camelCase)(node.name)}.ts`, render('definedTypesPage.njk', {
                imports: imports.toString({
                    ...dependencyMap,
                    generatedTypes: '.',
                }),
                typeWithCodecFragment,
                typeDiscriminatedUnionHelpersFragment,
            }));
        },
    }), (v) => (0, visitors_1.recordLinkablesVisitor)(v, linkables));
}
exports.getRenderMapVisitor = getRenderMapVisitor;
function getRenamedArgsMap(instruction) {
    var _a;
    const argNames = [
        ...instruction.arguments.map((a) => a.name),
        ...((_a = instruction.extraArguments) !== null && _a !== void 0 ? _a : []).map((a) => a.name),
    ];
    const duplicateArgs = argNames.filter((e, i, a) => a.indexOf(e) !== i);
    if (duplicateArgs.length > 0) {
        throw new Error(`Duplicate args found: [${duplicateArgs.join(', ')}] in instruction [${instruction.name}].`);
    }
    const allNames = [
        ...instruction.accounts.map((account) => account.name),
        ...argNames,
    ];
    const duplicates = allNames.filter((e, i, a) => a.indexOf(e) !== i);
    if (duplicates.length === 0)
        return new Map();
    (0, shared_1.logWarn)(`[JavaScriptExperimental] Accounts and args of instruction [${instruction.name}] have the following ` +
        `conflicting attributes [${duplicates.join(', ')}]. ` +
        `Thus, the arguments have been renamed to avoid conflicts in the input type.`);
    return new Map(duplicates.map((name) => [(0, shared_1.mainCase)(name), (0, shared_1.mainCase)(`${name}Arg`)]));
}
//# sourceMappingURL=getRenderMapVisitor.js.map