"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getRenderMapVisitor = void 0;
const sync_1 = require("@prettier/sync");
const nodes_1 = require("../../nodes");
const shared_1 = require("../../shared");
const visitors_1 = require("../../visitors");
const customDataHelpers_1 = require("./customDataHelpers");
const getTypeManifestVisitor_1 = require("./getTypeManifestVisitor");
const JavaScriptContextMap_1 = require("./JavaScriptContextMap");
const JavaScriptImportMap_1 = require("./JavaScriptImportMap");
const renderInstructionDefaults_1 = require("./renderInstructionDefaults");
const DEFAULT_PRETTIER_OPTIONS = {
    semi: true,
    singleQuote: true,
    trailingComma: 'es5',
    useTabs: false,
    tabWidth: 2,
    arrowParens: 'always',
    printWidth: 80,
    parser: 'typescript',
};
function getRenderMapVisitor(options = {}) {
    var _a, _b, _c, _d, _e, _f;
    const linkables = new shared_1.LinkableDictionary();
    const byteSizeVisitor = (0, visitors_1.getByteSizeVisitor)(linkables);
    let program = null;
    const renderParentInstructions = (_a = options.renderParentInstructions) !== null && _a !== void 0 ? _a : false;
    const formatCode = (_b = options.formatCode) !== null && _b !== void 0 ? _b : true;
    const prettierOptions = {
        ...DEFAULT_PRETTIER_OPTIONS,
        ...options.prettierOptions,
    };
    const dependencyMap = {
        generated: '..',
        hooked: '../../hooked',
        umi: '@metaplex-foundation/umi',
        umiSerializers: '@metaplex-foundation/umi/serializers',
        mplEssentials: '@metaplex-foundation/mpl-toolbox',
        mplToolbox: '@metaplex-foundation/mpl-toolbox',
        ...options.dependencyMap,
        // Custom relative dependencies to link generated files together.
        generatedPrograms: '../programs',
        generatedAccounts: '../accounts',
        generatedErrors: '../errors',
        generatedTypes: '../types',
    };
    const nonScalarEnums = ((_c = options.nonScalarEnums) !== null && _c !== void 0 ? _c : []).map(shared_1.mainCase);
    const internalNodes = ((_d = options.internalNodes) !== null && _d !== void 0 ? _d : []).map(shared_1.mainCase);
    const customAccountData = (0, customDataHelpers_1.parseCustomDataOptions)((_e = options.customAccountData) !== null && _e !== void 0 ? _e : [], 'AccountData');
    const customInstructionData = (0, customDataHelpers_1.parseCustomDataOptions)((_f = options.customInstructionData) !== null && _f !== void 0 ? _f : [], 'InstructionData');
    const getTypeManifestVisitor = (parentName) => (0, getTypeManifestVisitor_1.getTypeManifestVisitor)({
        linkables,
        nonScalarEnums,
        customAccountData,
        customInstructionData,
        parentName,
    });
    const typeManifestVisitor = getTypeManifestVisitor();
    const resolvedInstructionInputVisitor = (0, visitors_1.getResolvedInstructionInputsVisitor)();
    function getInstructionAccountType(account) {
        if (account.isPda && account.isSigner === false)
            return 'Pda';
        if (account.isSigner === 'either')
            return 'PublicKey | Pda | Signer';
        return account.isSigner ? 'Signer' : 'PublicKey | Pda';
    }
    function getInstructionAccountImports(accounts) {
        const imports = new JavaScriptImportMap_1.JavaScriptImportMap();
        accounts.forEach((account) => {
            if (account.isSigner !== true && !account.isPda)
                imports.add('umi', 'PublicKey');
            if (account.isSigner !== true)
                imports.add('umi', 'Pda');
            if (account.isSigner !== false)
                imports.add('umi', 'Signer');
        });
        return imports;
    }
    function getMergeConflictsForInstructionAccountsAndArgs(instruction) {
        var _a;
        const allNames = [
            ...instruction.accounts.map((account) => account.name),
            ...instruction.arguments.map((field) => field.name),
            ...((_a = instruction.extraArguments) !== null && _a !== void 0 ? _a : []).map((field) => field.name),
        ];
        const duplicates = allNames.filter((e, i, a) => a.indexOf(e) !== i);
        return [...new Set(duplicates)];
    }
    function render(template, context, renderOptions) {
        const code = (0, shared_1.resolveTemplate)(`${__dirname}/templates`, template, context, renderOptions);
        return formatCode ? (0, sync_1.format)(code, prettierOptions) : code;
    }
    return (0, shared_1.pipe)((0, visitors_1.staticVisitor)(() => new shared_1.RenderMap()), (v) => (0, visitors_1.extendVisitor)(v, {
        visitRoot(node, { self }) {
            const isNotInternal = (n) => !internalNodes.includes(n.name);
            const programsToExport = (0, nodes_1.getAllPrograms)(node).filter(isNotInternal);
            const accountsToExport = (0, nodes_1.getAllAccounts)(node).filter(isNotInternal);
            const instructionsToExport = (0, nodes_1.getAllInstructionsWithSubs)(node, {
                leavesOnly: !renderParentInstructions,
            }).filter(isNotInternal);
            const definedTypesToExport = (0, nodes_1.getAllDefinedTypes)(node).filter(isNotInternal);
            const hasAnythingToExport = programsToExport.length > 0 ||
                accountsToExport.length > 0 ||
                instructionsToExport.length > 0 ||
                definedTypesToExport.length > 0;
            const ctx = {
                root: node,
                programsToExport,
                accountsToExport,
                instructionsToExport,
                definedTypesToExport,
                hasAnythingToExport,
            };
            const map = new shared_1.RenderMap();
            if (hasAnythingToExport) {
                map.add('shared/index.ts', render('sharedPage.njk', ctx));
            }
            if (programsToExport.length > 0) {
                map
                    .add('programs/index.ts', render('programsIndex.njk', ctx))
                    .add('errors/index.ts', render('errorsIndex.njk', ctx));
            }
            if (accountsToExport.length > 0) {
                map.add('accounts/index.ts', render('accountsIndex.njk', ctx));
            }
            if (instructionsToExport.length > 0) {
                map.add('instructions/index.ts', render('instructionsIndex.njk', ctx));
            }
            if (definedTypesToExport.length > 0) {
                map.add('types/index.ts', render('definedTypesIndex.njk', ctx));
            }
            return map
                .add('index.ts', render('rootIndex.njk', ctx))
                .mergeWith(...(0, nodes_1.getAllPrograms)(node).map((p) => (0, visitors_1.visit)(p, self)));
        },
        visitProgram(node, { self }) {
            program = node;
            const pascalCaseName = (0, shared_1.pascalCase)(node.name);
            const customDataDefinedType = [
                ...(0, customDataHelpers_1.getDefinedTypeNodesToExtract)(node.accounts, customAccountData),
                ...(0, customDataHelpers_1.getDefinedTypeNodesToExtract)(node.instructions, customInstructionData),
            ];
            const renderMap = new shared_1.RenderMap()
                .mergeWith(...node.accounts.map((a) => (0, visitors_1.visit)(a, self)))
                .mergeWith(...node.definedTypes.map((t) => (0, visitors_1.visit)(t, self)))
                .mergeWith(...customDataDefinedType.map((t) => (0, visitors_1.visit)(t, self)))
                .mergeWith(...(0, nodes_1.getAllInstructionsWithSubs)(node, {
                leavesOnly: !renderParentInstructions,
            }).map((ix) => (0, visitors_1.visit)(ix, self)))
                .add(`errors/${(0, shared_1.camelCase)(node.name)}.ts`, render('errorsPage.njk', {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap()
                    .add('umi', ['ProgramError', 'Program'])
                    .toString(dependencyMap),
                program: node,
                errors: node.errors.map((error) => ({
                    ...error,
                    prefixedName: (0, shared_1.pascalCase)(node.prefix) + (0, shared_1.pascalCase)(error.name),
                })),
            }))
                .add(`programs/${(0, shared_1.camelCase)(node.name)}.ts`, render('programsPage.njk', {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap()
                    .add('umi', [
                    'ClusterFilter',
                    'Context',
                    'Program',
                    'PublicKey',
                ])
                    .add('errors', [
                    `get${pascalCaseName}ErrorFromCode`,
                    `get${pascalCaseName}ErrorFromName`,
                ])
                    .toString(dependencyMap),
                program: node,
            }));
            program = null;
            return renderMap;
        },
        visitAccount(node) {
            var _a, _b, _c;
            const customData = customAccountData.get(node.name);
            const isLinked = !!customData;
            const typeManifest = (0, visitors_1.visit)(node, typeManifestVisitor);
            const imports = new JavaScriptImportMap_1.JavaScriptImportMap().mergeWith(typeManifest.strictImports, typeManifest.serializerImports);
            if (!isLinked) {
                imports.mergeWith(typeManifest.looseImports);
            }
            imports
                .add('umi', [
                'Account',
                'assertAccountExists',
                'Context',
                'deserializeAccount',
                'Pda',
                'PublicKey',
                'publicKey',
                'RpcAccount',
                'RpcGetAccountOptions',
                'RpcGetAccountsOptions',
            ])
                .add('umiSerializers', !isLinked ? ['Serializer'] : [])
                .addAlias('umi', 'publicKey', 'toPublicKey');
            // Discriminator.
            const discriminator = (_b = ((_a = node.discriminators) !== null && _a !== void 0 ? _a : []).find((d) => !(0, nodes_1.isNode)(d, 'constantDiscriminatorNode'))) !== null && _b !== void 0 ? _b : null;
            let resolvedDiscriminator = null;
            if ((0, nodes_1.isNode)(discriminator, 'fieldDiscriminatorNode')) {
                const discriminatorField = (0, nodes_1.resolveNestedTypeNode)(node.data).fields.find((f) => f.name === discriminator.name);
                const discriminatorValue = (discriminatorField === null || discriminatorField === void 0 ? void 0 : discriminatorField.defaultValue)
                    ? (0, visitors_1.visit)(discriminatorField.defaultValue, typeManifestVisitor)
                    : undefined;
                if (discriminatorValue) {
                    imports.mergeWith(discriminatorValue.valueImports);
                    resolvedDiscriminator = {
                        ...discriminator,
                        value: discriminatorValue.value,
                    };
                }
            }
            else if ((0, nodes_1.isNode)(discriminator, 'sizeDiscriminatorNode')) {
                resolvedDiscriminator = discriminator;
            }
            // GPA Fields.
            const gpaFields = (0, shared_1.getGpaFieldsFromAccount)(node, byteSizeVisitor).map((gpaField) => {
                const gpaFieldManifest = (0, visitors_1.visit)(gpaField.type, typeManifestVisitor);
                imports.mergeWith(gpaFieldManifest.looseImports, gpaFieldManifest.serializerImports);
                return { ...gpaField, manifest: gpaFieldManifest };
            });
            let resolvedGpaFields = null;
            if (gpaFields.length > 0) {
                imports.add('umi', ['gpaBuilder']);
                resolvedGpaFields = {
                    type: `{ ${gpaFields
                        .map((f) => `'${f.name}': ${f.manifest.looseType}`)
                        .join(', ')} }`,
                    argument: `{ ${gpaFields
                        .map((f) => {
                        const offset = f.offset === null ? 'null' : `${f.offset}`;
                        return `'${f.name}': [${offset}, ${f.manifest.serializer}]`;
                    })
                        .join(', ')} }`,
                };
            }
            // Seeds.
            const pda = node.pda ? linkables.get(node.pda) : undefined;
            const pdaSeeds = (_c = pda === null || pda === void 0 ? void 0 : pda.seeds) !== null && _c !== void 0 ? _c : [];
            const seeds = pdaSeeds.map((seed) => {
                if ((0, nodes_1.isNode)(seed, 'variablePdaSeedNode')) {
                    const seedManifest = (0, visitors_1.visit)(seed.type, typeManifestVisitor);
                    imports.mergeWith(seedManifest.looseImports, seedManifest.serializerImports);
                    return { ...seed, typeManifest: seedManifest };
                }
                if ((0, nodes_1.isNode)(seed.value, 'programIdValueNode')) {
                    imports
                        .add('umiSerializers', 'publicKey')
                        .addAlias('umiSerializers', 'publicKey', 'publicKeySerializer');
                    return seed;
                }
                const seedManifest = (0, visitors_1.visit)(seed.type, typeManifestVisitor);
                imports.mergeWith(seedManifest.serializerImports);
                const seedValue = seed.value;
                const valueManifest = (0, visitors_1.visit)(seedValue, typeManifestVisitor);
                seedValue.render = valueManifest.value;
                imports.mergeWith(valueManifest.valueImports);
                return { ...seed, typeManifest: seedManifest };
            });
            if (seeds.length > 0) {
                imports.add('umi', ['Pda']);
            }
            const hasVariableSeeds = pdaSeeds.filter((0, nodes_1.isNodeFilter)('variablePdaSeedNode')).length > 0;
            return new shared_1.RenderMap().add(`accounts/${(0, shared_1.camelCase)(node.name)}.ts`, render('accountsPage.njk', {
                account: node,
                imports: imports.toString(dependencyMap),
                program,
                typeManifest,
                discriminator: resolvedDiscriminator,
                gpaFields: resolvedGpaFields,
                seeds,
                hasVariableSeeds,
                customData,
            }));
        },
        visitInstruction(node) {
            var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
            // Imports and interfaces.
            const interfaces = new JavaScriptContextMap_1.JavaScriptContextMap().add('programs');
            const imports = new JavaScriptImportMap_1.JavaScriptImportMap()
                .add('umi', ['Context', 'TransactionBuilder', 'transactionBuilder'])
                .add('shared', [
                'ResolvedAccount',
                'ResolvedAccountsWithIndices',
                'getAccountMetasAndSigners',
            ]);
            // Instruction helpers.
            const customData = customInstructionData.get(node.name);
            const linkedDataArgs = !!customData;
            const hasAccounts = node.accounts.length > 0;
            const hasData = linkedDataArgs || node.arguments.length > 0;
            const hasDataArgs = linkedDataArgs ||
                node.arguments.filter((field) => field.defaultValueStrategy !== 'omitted').length > 0;
            const hasExtraArgs = ((_a = node.extraArguments) !== null && _a !== void 0 ? _a : []).filter((field) => field.defaultValueStrategy !== 'omitted').length > 0;
            const hasAnyArgs = hasDataArgs || hasExtraArgs;
            const allArgumentsWithDefaultValue = [
                ...node.arguments.filter((a) => a.defaultValue && !(0, nodes_1.isNode)(a.defaultValue, nodes_1.VALUE_NODES)),
                ...((_b = node.extraArguments) !== null && _b !== void 0 ? _b : []).filter((a) => a.defaultValue),
            ];
            const hasArgDefaults = allArgumentsWithDefaultValue.length > 0;
            const hasArgResolvers = allArgumentsWithDefaultValue.some((a) => (0, nodes_1.isNode)(a.defaultValue, 'resolverValueNode'));
            const hasAccountResolvers = node.accounts.some((a) => (0, nodes_1.isNode)(a.defaultValue, 'resolverValueNode'));
            const byteDelta = (_d = (_c = node.byteDeltas) === null || _c === void 0 ? void 0 : _c[0]) !== null && _d !== void 0 ? _d : undefined;
            const hasByteResolver = byteDelta && (0, nodes_1.isNode)(byteDelta.value, 'resolverValueNode');
            let remainingAccounts = (_f = (_e = node.remainingAccounts) === null || _e === void 0 ? void 0 : _e[0]) !== null && _f !== void 0 ? _f : undefined;
            if (remainingAccounts &&
                (0, nodes_1.isNode)(remainingAccounts.value, 'argumentValueNode') &&
                (0, nodes_1.getAllInstructionArguments)(node).every((arg) => arg.name !== (remainingAccounts === null || remainingAccounts === void 0 ? void 0 : remainingAccounts.value.name))) {
                remainingAccounts = undefined;
            }
            const hasRemainingAccountsResolver = remainingAccounts &&
                (0, nodes_1.isNode)(remainingAccounts.value, 'resolverValueNode');
            const hasResolvers = hasArgResolvers ||
                hasAccountResolvers ||
                hasByteResolver ||
                hasRemainingAccountsResolver;
            const hasResolvedArgs = hasDataArgs || hasArgDefaults || hasResolvers;
            if (hasResolvers) {
                interfaces.add(['eddsa', 'identity', 'payer']);
            }
            // canMergeAccountsAndArgs
            let canMergeAccountsAndArgs = false;
            if (!linkedDataArgs) {
                const accountsAndArgsConflicts = getMergeConflictsForInstructionAccountsAndArgs(node);
                if (accountsAndArgsConflicts.length > 0) {
                    (0, shared_1.logWarn)(`[JavaScript] Accounts and args of instruction [${node.name}] have the following ` +
                        `conflicting attributes [${accountsAndArgsConflicts.join(', ')}]. ` +
                        `Thus, they could not be merged into a single input object. ` +
                        'You may want to rename the conflicting attributes.');
                }
                canMergeAccountsAndArgs = accountsAndArgsConflicts.length === 0;
            }
            // Resolved inputs.
            let argObject = canMergeAccountsAndArgs ? 'input' : 'args';
            argObject = hasResolvedArgs ? 'resolvedArgs' : argObject;
            const resolvedInputs = (0, visitors_1.visit)(node, resolvedInstructionInputVisitor).map((input) => {
                const renderedInput = (0, renderInstructionDefaults_1.renderInstructionDefaults)(input, typeManifestVisitor, node.optionalAccountStrategy, argObject);
                imports.mergeWith(renderedInput.imports);
                interfaces.mergeWith(renderedInput.interfaces);
                return { ...input, render: renderedInput.render };
            });
            const resolvedInputsWithDefaults = resolvedInputs.filter((input) => input.defaultValue !== undefined && input.render !== '');
            const argsWithDefaults = resolvedInputsWithDefaults
                .filter((0, nodes_1.isNodeFilter)('instructionArgumentNode'))
                .map((input) => input.name);
            // Accounts.
            const accounts = node.accounts.map((account) => {
                const hasDefaultValue = !!account.defaultValue;
                const resolvedAccount = resolvedInputs.find((input) => input.kind === 'instructionAccountNode' &&
                    input.name === account.name);
                return {
                    ...resolvedAccount,
                    type: getInstructionAccountType(resolvedAccount),
                    optionalSign: hasDefaultValue || account.isOptional ? '?' : '',
                    hasDefaultValue,
                };
            });
            imports.mergeWith(getInstructionAccountImports(accounts));
            // Data Args.
            const dataArgManifest = (0, visitors_1.visit)(node, typeManifestVisitor);
            if (linkedDataArgs || hasData) {
                imports.mergeWith(dataArgManifest.looseImports, dataArgManifest.serializerImports);
            }
            if (!linkedDataArgs) {
                imports.mergeWith(dataArgManifest.strictImports);
            }
            if (!linkedDataArgs && hasData) {
                imports.add('umiSerializers', ['Serializer']);
            }
            // Extra args.
            const extraArgStruct = (0, nodes_1.structTypeNodeFromInstructionArgumentNodes)((_g = node.extraArguments) !== null && _g !== void 0 ? _g : []);
            const visitor = getTypeManifestVisitor({
                strict: `${node.name}InstructionExtra`,
                loose: `${node.name}InstructionExtraArgs`,
            });
            const extraArgManifest = (0, visitors_1.visit)(extraArgStruct, visitor);
            imports.mergeWith(extraArgManifest.looseImports);
            // Arg defaults.
            allArgumentsWithDefaultValue.forEach((argument) => {
                var _a;
                if ((0, nodes_1.isNode)(argument.defaultValue, 'resolverValueNode')) {
                    imports.add((_a = argument.defaultValue.importFrom) !== null && _a !== void 0 ? _a : 'hooked', (0, shared_1.camelCase)(argument.defaultValue.name));
                }
            });
            if (argsWithDefaults.length > 0) {
                imports.add('shared', ['PickPartial']);
            }
            // Bytes created on chain.
            if (byteDelta && byteDelta.withHeader) {
                imports.add('umi', 'ACCOUNT_HEADER_SIZE');
            }
            if (byteDelta && (0, nodes_1.isNode)(byteDelta.value, 'accountLinkNode')) {
                const accountName = (0, shared_1.pascalCase)(byteDelta.value.name);
                const importFrom = (_h = byteDelta.value.importFrom) !== null && _h !== void 0 ? _h : 'generatedAccounts';
                imports.add(importFrom, `get${accountName}Size`);
            }
            else if (byteDelta &&
                (0, nodes_1.isNode)(byteDelta.value, 'resolverValueNode')) {
                imports.add((_j = byteDelta.value.importFrom) !== null && _j !== void 0 ? _j : 'hooked', (0, shared_1.camelCase)(byteDelta.value.name));
            }
            // Remaining accounts.
            if (remainingAccounts &&
                (0, nodes_1.isNode)(remainingAccounts.value, 'resolverValueNode')) {
                imports.add((_k = remainingAccounts.value.importFrom) !== null && _k !== void 0 ? _k : 'hooked', (0, shared_1.camelCase)(remainingAccounts.value.name));
            }
            return new shared_1.RenderMap().add(`instructions/${(0, shared_1.camelCase)(node.name)}.ts`, render('instructionsPage.njk', {
                instruction: node,
                imports: imports.toString(dependencyMap),
                interfaces: interfaces.toString(),
                program,
                resolvedInputs,
                resolvedInputsWithDefaults,
                argsWithDefaults,
                accounts,
                dataArgManifest,
                extraArgManifest,
                canMergeAccountsAndArgs,
                hasAccounts,
                hasData,
                hasDataArgs,
                hasExtraArgs,
                hasAnyArgs,
                hasArgDefaults,
                hasArgResolvers,
                hasAccountResolvers,
                hasByteResolver,
                hasRemainingAccountsResolver,
                hasResolvers,
                hasResolvedArgs,
                customData,
                remainingAccounts,
                byteDelta,
            }));
        },
        visitDefinedType(node) {
            const pascalCaseName = (0, shared_1.pascalCase)(node.name);
            const typeManifest = (0, visitors_1.visit)(node, typeManifestVisitor);
            const imports = new JavaScriptImportMap_1.JavaScriptImportMap()
                .mergeWithManifest(typeManifest)
                .add('umiSerializers', ['Serializer'])
                .remove('generatedTypes', [
                pascalCaseName,
                `${pascalCaseName}Args`,
                `get${pascalCaseName}Serializer`,
            ]);
            return new shared_1.RenderMap().add(`types/${(0, shared_1.camelCase)(node.name)}.ts`, render('definedTypesPage.njk', {
                definedType: node,
                imports: imports.toString({
                    ...dependencyMap,
                    generatedTypes: '.',
                }),
                typeManifest,
                isDataEnum: (0, nodes_1.isNode)(node.type, 'enumTypeNode') && (0, nodes_1.isDataEnum)(node.type),
            }));
        },
    }), (v) => (0, visitors_1.recordLinkablesVisitor)(v, linkables));
}
exports.getRenderMapVisitor = getRenderMapVisitor;
//# sourceMappingURL=getRenderMapVisitor.js.map