"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDebugStringVisitor = void 0;
const shared_1 = require("../shared");
const interceptVisitor_1 = require("./interceptVisitor");
const mergeVisitor_1 = require("./mergeVisitor");
function getDebugStringVisitor(options = {}) {
    var _a, _b;
    const indent = (_a = options.indent) !== null && _a !== void 0 ? _a : false;
    const indentSeparator = (_b = options.indentSeparator) !== null && _b !== void 0 ? _b : '|   ';
    let stackLevel = -1;
    return (0, shared_1.pipe)((0, mergeVisitor_1.mergeVisitor)((node) => {
        const details = getNodeDetails(node).join('.');
        if (indent) {
            return `${indentSeparator.repeat(stackLevel)}${node.kind}${details ? ` [${details}]` : ''}`;
        }
        return `${node.kind}${details ? `[${details}]` : ''}`;
    }, (node, values) => {
        const details = getNodeDetails(node).join('.');
        if (indent) {
            return [
                `${indentSeparator.repeat(stackLevel)}${node.kind}${details ? ` [${details}]` : ''}`,
                ...values,
            ].join('\n');
        }
        return `${node.kind}${details ? `[${details}]` : ''}(${values.join(', ')})`;
    }), (v) => (0, interceptVisitor_1.interceptVisitor)(v, (node, next) => {
        stackLevel += 1;
        const newNode = next(node);
        stackLevel -= 1;
        return newNode;
    }));
}
exports.getDebugStringVisitor = getDebugStringVisitor;
function getNodeDetails(node) {
    var _a, _b;
    switch (node.kind) {
        case 'programNode':
            return [node.name, node.publicKey];
        case 'instructionAccountNode':
            return [
                node.name,
                ...(node.isWritable ? ['writable'] : []),
                ...(node.isSigner === true ? ['signer'] : []),
                ...(node.isSigner === 'either' ? ['optionalSigner'] : []),
                ...(node.isOptional ? ['optional'] : []),
            ];
        case 'instructionRemainingAccountsNode':
            return [
                ...(node.isOptional ? ['optional'] : []),
                ...(node.isWritable ? ['writable'] : []),
                ...(node.isSigner === true ? ['signer'] : []),
                ...(node.isSigner === 'either' ? ['optionalSigner'] : []),
            ];
        case 'instructionByteDeltaNode':
            return [
                ...(node.subtract ? ['subtract'] : []),
                ...(node.withHeader ? ['withHeader'] : []),
            ];
        case 'errorNode':
            return [node.code.toString(), node.name];
        case 'programLinkNode':
        case 'pdaLinkNode':
        case 'accountLinkNode':
        case 'definedTypeLinkNode':
            return [
                node.name,
                ...(node.importFrom ? [`from:${node.importFrom}`] : []),
            ];
        case 'numberTypeNode':
            return [node.format, ...(node.endian === 'be' ? ['bigEndian'] : [])];
        case 'amountTypeNode':
            return [node.decimals.toString(), ...(node.unit ? [node.unit] : [])];
        case 'stringTypeNode':
            return [node.encoding];
        case 'optionTypeNode':
            return node.fixed ? ['fixed'] : [];
        case 'fixedCountNode':
            return [node.value.toString()];
        case 'numberValueNode':
            return [node.number.toString()];
        case 'stringValueNode':
            return [node.string];
        case 'booleanValueNode':
            return [node.boolean ? 'true' : 'false'];
        case 'bytesValueNode':
            return [node.encoding, node.data];
        case 'publicKeyValueNode':
            return [
                ...(node.identifier ? [`${node.identifier}`] : []),
                node.publicKey,
            ];
        case 'enumValueNode':
            return [node.variant];
        case 'resolverValueNode':
            return [
                node.name,
                ...(node.importFrom ? [`from:${node.importFrom}`] : []),
            ];
        case 'constantDiscriminatorNode':
            return [...(node.offset > 0 ? [`offset:${node.offset}`] : [])];
        case 'fieldDiscriminatorNode':
            return [node.name, ...(node.offset > 0 ? [`offset:${node.offset}`] : [])];
        case 'sizeDiscriminatorNode':
            return [node.size.toString()];
        case 'fixedSizeTypeNode':
            return [node.size.toString()];
        case 'preOffsetTypeNode':
            return [node.offset.toString(), (_a = node.strategy) !== null && _a !== void 0 ? _a : 'relative'];
        case 'postOffsetTypeNode':
            return [node.offset.toString(), (_b = node.strategy) !== null && _b !== void 0 ? _b : 'relative'];
        default:
            return 'name' in node ? [node.name] : [];
    }
}
//# sourceMappingURL=getDebugStringVisitor.js.map