"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.identityVisitor = void 0;
const nodes_1 = require("../nodes");
const staticVisitor_1 = require("./staticVisitor");
const visitor_1 = require("./visitor");
function identityVisitor(nodeKeys = nodes_1.REGISTERED_NODE_KINDS) {
    const castedNodeKeys = nodeKeys;
    const visitor = (0, staticVisitor_1.staticVisitor)((node) => ({ ...node }), castedNodeKeys);
    const visit = (v) => (node) => castedNodeKeys.includes(node.kind) ? (0, visitor_1.visit)(node, v) : { ...node };
    if (castedNodeKeys.includes('rootNode')) {
        visitor.visitRoot = function visitRoot(node) {
            const program = visit(this)(node.program);
            if (program === null)
                return null;
            (0, nodes_1.assertIsNode)(program, 'programNode');
            return (0, nodes_1.rootNode)(program, node.additionalPrograms
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('programNode')));
        };
    }
    if (castedNodeKeys.includes('programNode')) {
        visitor.visitProgram = function visitProgram(node) {
            return (0, nodes_1.programNode)({
                ...node,
                pdas: node.pdas
                    .map(visit(this))
                    .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('pdaNode')),
                accounts: node.accounts
                    .map(visit(this))
                    .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('accountNode')),
                instructions: node.instructions
                    .map(visit(this))
                    .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('instructionNode')),
                definedTypes: node.definedTypes
                    .map(visit(this))
                    .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('definedTypeNode')),
                errors: node.errors
                    .map(visit(this))
                    .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('errorNode')),
            });
        };
    }
    if (castedNodeKeys.includes('pdaNode')) {
        visitor.visitPda = function visitPda(node) {
            return (0, nodes_1.pdaNode)(node.name, node.seeds
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)(nodes_1.PDA_SEED_NODES)));
        };
    }
    if (castedNodeKeys.includes('accountNode')) {
        visitor.visitAccount = function visitAccount(node) {
            var _a;
            const data = visit(this)(node.data);
            if (data === null)
                return null;
            (0, nodes_1.assertIsNode)(data, 'structTypeNode');
            const pda = node.pda ? (_a = visit(this)(node.pda)) !== null && _a !== void 0 ? _a : undefined : undefined;
            if (pda)
                (0, nodes_1.assertIsNode)(pda, 'pdaLinkNode');
            return (0, nodes_1.accountNode)({ ...node, data, pda });
        };
    }
    if (castedNodeKeys.includes('instructionNode')) {
        visitor.visitInstruction = function visitInstruction(node) {
            return (0, nodes_1.instructionNode)({
                ...node,
                accounts: node.accounts
                    .map(visit(this))
                    .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('instructionAccountNode')),
                arguments: node.arguments
                    .map(visit(this))
                    .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('instructionArgumentNode')),
                extraArguments: node.extraArguments
                    ? node.extraArguments
                        .map(visit(this))
                        .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('instructionArgumentNode'))
                    : undefined,
                remainingAccounts: node.remainingAccounts
                    ? node.remainingAccounts
                        .map(visit(this))
                        .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('instructionRemainingAccountsNode'))
                    : undefined,
                byteDeltas: node.byteDeltas
                    ? node.byteDeltas
                        .map(visit(this))
                        .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('instructionByteDeltaNode'))
                    : undefined,
                discriminators: node.discriminators
                    ? node.discriminators
                        .map(visit(this))
                        .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)(nodes_1.DISCRIMINATOR_NODES))
                    : undefined,
                subInstructions: node.subInstructions
                    ? node.subInstructions
                        .map(visit(this))
                        .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('instructionNode'))
                    : undefined,
            });
        };
    }
    if (castedNodeKeys.includes('instructionAccountNode')) {
        visitor.visitInstructionAccount = function visitInstructionAccount(node) {
            var _a;
            const defaultValue = node.defaultValue
                ? (_a = visit(this)(node.defaultValue)) !== null && _a !== void 0 ? _a : undefined
                : undefined;
            if (defaultValue)
                (0, nodes_1.assertIsNode)(defaultValue, nodes_1.INSTRUCTION_INPUT_VALUE_NODE);
            return (0, nodes_1.instructionAccountNode)({ ...node, defaultValue });
        };
    }
    if (castedNodeKeys.includes('instructionArgumentNode')) {
        visitor.visitInstructionArgument = function visitInstructionArgument(node) {
            var _a;
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            const defaultValue = node.defaultValue
                ? (_a = visit(this)(node.defaultValue)) !== null && _a !== void 0 ? _a : undefined
                : undefined;
            if (defaultValue)
                (0, nodes_1.assertIsNode)(defaultValue, nodes_1.INSTRUCTION_INPUT_VALUE_NODE);
            return (0, nodes_1.instructionArgumentNode)({ ...node, type, defaultValue });
        };
    }
    if (castedNodeKeys.includes('instructionRemainingAccountsNode')) {
        visitor.visitInstructionRemainingAccounts =
            function visitInstructionRemainingAccounts(node) {
                const value = visit(this)(node.value);
                if (value === null)
                    return null;
                (0, nodes_1.assertIsNode)(value, ['argumentValueNode', 'resolverValueNode']);
                return (0, nodes_1.instructionRemainingAccountsNode)(value, { ...node });
            };
    }
    if (castedNodeKeys.includes('instructionByteDeltaNode')) {
        visitor.visitInstructionByteDelta = function visitInstructionByteDelta(node) {
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, [
                'numberValueNode',
                'accountLinkNode',
                'argumentValueNode',
                'resolverValueNode',
            ]);
            return (0, nodes_1.instructionByteDeltaNode)(value, { ...node });
        };
    }
    if (castedNodeKeys.includes('definedTypeNode')) {
        visitor.visitDefinedType = function visitDefinedType(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.definedTypeNode)({ ...node, type });
        };
    }
    if (castedNodeKeys.includes('arrayTypeNode')) {
        visitor.visitArrayType = function visitArrayType(node) {
            const size = visit(this)(node.count);
            if (size === null)
                return null;
            (0, nodes_1.assertIsNode)(size, nodes_1.COUNT_NODES);
            const item = visit(this)(node.item);
            if (item === null)
                return null;
            (0, nodes_1.assertIsNode)(item, nodes_1.TYPE_NODES);
            return (0, nodes_1.arrayTypeNode)(item, size);
        };
    }
    if (castedNodeKeys.includes('enumTypeNode')) {
        visitor.visitEnumType = function visitEnumType(node) {
            return (0, nodes_1.enumTypeNode)(node.variants
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)(nodes_1.ENUM_VARIANT_TYPE_NODES)), { ...node });
        };
    }
    if (castedNodeKeys.includes('enumStructVariantTypeNode')) {
        visitor.visitEnumStructVariantType = function visitEnumStructVariantType(node) {
            const newStruct = visit(this)(node.struct);
            if (!newStruct) {
                return (0, nodes_1.enumEmptyVariantTypeNode)(node.name);
            }
            (0, nodes_1.assertIsNode)(newStruct, 'structTypeNode');
            if (newStruct.fields.length === 0) {
                return (0, nodes_1.enumEmptyVariantTypeNode)(node.name);
            }
            return (0, nodes_1.enumStructVariantTypeNode)(node.name, newStruct);
        };
    }
    if (castedNodeKeys.includes('enumTupleVariantTypeNode')) {
        visitor.visitEnumTupleVariantType = function visitEnumTupleVariantType(node) {
            const newTuple = visit(this)(node.tuple);
            if (!newTuple) {
                return (0, nodes_1.enumEmptyVariantTypeNode)(node.name);
            }
            (0, nodes_1.assertIsNode)(newTuple, 'tupleTypeNode');
            if (newTuple.items.length === 0) {
                return (0, nodes_1.enumEmptyVariantTypeNode)(node.name);
            }
            return (0, nodes_1.enumTupleVariantTypeNode)(node.name, newTuple);
        };
    }
    if (castedNodeKeys.includes('mapTypeNode')) {
        visitor.visitMapType = function visitMapType(node) {
            const size = visit(this)(node.count);
            if (size === null)
                return null;
            (0, nodes_1.assertIsNode)(size, nodes_1.COUNT_NODES);
            const key = visit(this)(node.key);
            if (key === null)
                return null;
            (0, nodes_1.assertIsNode)(key, nodes_1.TYPE_NODES);
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, nodes_1.TYPE_NODES);
            return (0, nodes_1.mapTypeNode)(key, value, size, node.idlMap);
        };
    }
    if (castedNodeKeys.includes('optionTypeNode')) {
        visitor.visitOptionType = function visitOptionType(node) {
            const prefix = visit(this)(node.prefix);
            if (prefix === null)
                return null;
            (0, nodes_1.assertIsNestedTypeNode)(prefix, 'numberTypeNode');
            const item = visit(this)(node.item);
            if (item === null)
                return null;
            (0, nodes_1.assertIsNode)(item, nodes_1.TYPE_NODES);
            return (0, nodes_1.optionTypeNode)(item, { ...node, prefix });
        };
    }
    if (castedNodeKeys.includes('zeroableOptionTypeNode')) {
        visitor.visitZeroableOptionType = function visitZeroableOptionType(node) {
            var _a;
            const item = visit(this)(node.item);
            if (item === null)
                return null;
            (0, nodes_1.assertIsNode)(item, nodes_1.TYPE_NODES);
            const zeroValue = node.zeroValue
                ? (_a = visit(this)(node.zeroValue)) !== null && _a !== void 0 ? _a : undefined
                : undefined;
            if (zeroValue)
                (0, nodes_1.assertIsNode)(zeroValue, 'constantValueNode');
            return (0, nodes_1.zeroableOptionTypeNode)(item, zeroValue);
        };
    }
    if (castedNodeKeys.includes('booleanTypeNode')) {
        visitor.visitBooleanType = function visitBooleanType(node) {
            const size = visit(this)(node.size);
            if (size === null)
                return null;
            (0, nodes_1.assertIsNestedTypeNode)(size, 'numberTypeNode');
            return (0, nodes_1.booleanTypeNode)(size);
        };
    }
    if (castedNodeKeys.includes('setTypeNode')) {
        visitor.visitSetType = function visitSetType(node) {
            const size = visit(this)(node.count);
            if (size === null)
                return null;
            (0, nodes_1.assertIsNode)(size, nodes_1.COUNT_NODES);
            const item = visit(this)(node.item);
            if (item === null)
                return null;
            (0, nodes_1.assertIsNode)(item, nodes_1.TYPE_NODES);
            return (0, nodes_1.setTypeNode)(item, size, node.idlSet);
        };
    }
    if (castedNodeKeys.includes('structTypeNode')) {
        visitor.visitStructType = function visitStructType(node) {
            const fields = node.fields
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('structFieldTypeNode'));
            return (0, nodes_1.structTypeNode)(fields);
        };
    }
    if (castedNodeKeys.includes('structFieldTypeNode')) {
        visitor.visitStructFieldType = function visitStructFieldType(node) {
            var _a;
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            const defaultValue = node.defaultValue
                ? (_a = visit(this)(node.defaultValue)) !== null && _a !== void 0 ? _a : undefined
                : undefined;
            if (defaultValue)
                (0, nodes_1.assertIsNode)(defaultValue, nodes_1.VALUE_NODES);
            return (0, nodes_1.structFieldTypeNode)({ ...node, type, defaultValue });
        };
    }
    if (castedNodeKeys.includes('tupleTypeNode')) {
        visitor.visitTupleType = function visitTupleType(node) {
            const items = node.items
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)(nodes_1.TYPE_NODES));
            return (0, nodes_1.tupleTypeNode)(items);
        };
    }
    if (castedNodeKeys.includes('amountTypeNode')) {
        visitor.visitAmountType = function visitAmountType(node) {
            const number = visit(this)(node.number);
            if (number === null)
                return null;
            (0, nodes_1.assertIsNestedTypeNode)(number, 'numberTypeNode');
            return (0, nodes_1.amountTypeNode)(number, node.decimals, node.unit);
        };
    }
    if (castedNodeKeys.includes('dateTimeTypeNode')) {
        visitor.visitDateTimeType = function visitDateTimeType(node) {
            const number = visit(this)(node.number);
            if (number === null)
                return null;
            (0, nodes_1.assertIsNestedTypeNode)(number, 'numberTypeNode');
            return (0, nodes_1.dateTimeTypeNode)(number);
        };
    }
    if (castedNodeKeys.includes('solAmountTypeNode')) {
        visitor.visitSolAmountType = function visitSolAmountType(node) {
            const number = visit(this)(node.number);
            if (number === null)
                return null;
            (0, nodes_1.assertIsNestedTypeNode)(number, 'numberTypeNode');
            return (0, nodes_1.solAmountTypeNode)(number);
        };
    }
    if (castedNodeKeys.includes('prefixedCountNode')) {
        visitor.visitPrefixedCount = function visitPrefixedCount(node) {
            const prefix = visit(this)(node.prefix);
            if (prefix === null)
                return null;
            (0, nodes_1.assertIsNestedTypeNode)(prefix, 'numberTypeNode');
            return (0, nodes_1.prefixedCountNode)(prefix);
        };
    }
    if (castedNodeKeys.includes('arrayValueNode')) {
        visitor.visitArrayValue = function visitArrayValue(node) {
            return (0, nodes_1.arrayValueNode)(node.items
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)(nodes_1.VALUE_NODES)));
        };
    }
    if (castedNodeKeys.includes('constantValueNode')) {
        visitor.visitConstantValue = function visitConstantValue(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, nodes_1.VALUE_NODES);
            return (0, nodes_1.constantValueNode)(type, value);
        };
    }
    if (castedNodeKeys.includes('enumValueNode')) {
        visitor.visitEnumValue = function visitEnumValue(node) {
            var _a;
            const enumLink = visit(this)(node.enum);
            if (enumLink === null)
                return null;
            (0, nodes_1.assertIsNode)(enumLink, ['definedTypeLinkNode']);
            if (typeof node.value === 'string')
                return { ...node };
            const value = node.value
                ? (_a = visit(this)(node.value)) !== null && _a !== void 0 ? _a : undefined
                : undefined;
            if (value)
                (0, nodes_1.assertIsNode)(value, ['structValueNode', 'tupleValueNode']);
            return (0, nodes_1.enumValueNode)(enumLink, node.variant, value);
        };
    }
    if (castedNodeKeys.includes('mapValueNode')) {
        visitor.visitMapValue = function visitMapValue(node) {
            return (0, nodes_1.mapValueNode)(node.entries
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('mapEntryValueNode')));
        };
    }
    if (castedNodeKeys.includes('mapEntryValueNode')) {
        visitor.visitMapEntryValue = function visitMapEntryValue(node) {
            const key = visit(this)(node.key);
            if (key === null)
                return null;
            (0, nodes_1.assertIsNode)(key, nodes_1.VALUE_NODES);
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, nodes_1.VALUE_NODES);
            return (0, nodes_1.mapEntryValueNode)(key, value);
        };
    }
    if (castedNodeKeys.includes('setValueNode')) {
        visitor.visitSetValue = function visitSetValue(node) {
            return (0, nodes_1.setValueNode)(node.items
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)(nodes_1.VALUE_NODES)));
        };
    }
    if (castedNodeKeys.includes('someValueNode')) {
        visitor.visitSomeValue = function visitSomeValue(node) {
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, nodes_1.VALUE_NODES);
            return (0, nodes_1.someValueNode)(value);
        };
    }
    if (castedNodeKeys.includes('structValueNode')) {
        visitor.visitStructValue = function visitStructValue(node) {
            return (0, nodes_1.structValueNode)(node.fields
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('structFieldValueNode')));
        };
    }
    if (castedNodeKeys.includes('structFieldValueNode')) {
        visitor.visitStructFieldValue = function visitStructFieldValue(node) {
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, nodes_1.VALUE_NODES);
            return (0, nodes_1.structFieldValueNode)(node.name, value);
        };
    }
    if (castedNodeKeys.includes('tupleValueNode')) {
        visitor.visitTupleValue = function visitTupleValue(node) {
            return (0, nodes_1.tupleValueNode)(node.items
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)(nodes_1.VALUE_NODES)));
        };
    }
    if (castedNodeKeys.includes('constantPdaSeedNode')) {
        visitor.visitConstantPdaSeed = function visitConstantPdaSeed(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, [...nodes_1.VALUE_NODES, 'programIdValueNode']);
            return (0, nodes_1.constantPdaSeedNode)(type, value);
        };
    }
    if (castedNodeKeys.includes('variablePdaSeedNode')) {
        visitor.visitVariablePdaSeed = function visitVariablePdaSeed(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.variablePdaSeedNode)(node.name, type, node.docs);
        };
    }
    if (castedNodeKeys.includes('resolverValueNode')) {
        visitor.visitResolverValue = function visitResolverValue(node) {
            var _a;
            const dependsOn = ((_a = node.dependsOn) !== null && _a !== void 0 ? _a : [])
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)([
                'accountValueNode',
                'argumentValueNode',
            ]));
            return (0, nodes_1.resolverValueNode)(node.name, {
                ...node,
                dependsOn: dependsOn.length === 0 ? undefined : dependsOn,
            });
        };
    }
    if (castedNodeKeys.includes('conditionalValueNode')) {
        visitor.visitConditionalValue = function visitConditionalValue(node) {
            var _a, _b, _c;
            const condition = visit(this)(node.condition);
            if (condition === null)
                return null;
            (0, nodes_1.assertIsNode)(condition, [
                'resolverValueNode',
                'accountValueNode',
                'argumentValueNode',
            ]);
            const value = node.value
                ? (_a = visit(this)(node.value)) !== null && _a !== void 0 ? _a : undefined
                : undefined;
            if (value)
                (0, nodes_1.assertIsNode)(value, nodes_1.VALUE_NODES);
            const ifTrue = node.ifTrue
                ? (_b = visit(this)(node.ifTrue)) !== null && _b !== void 0 ? _b : undefined
                : undefined;
            if (ifTrue)
                (0, nodes_1.assertIsNode)(ifTrue, nodes_1.CONDITIONAL_VALUE_BRANCH_NODES);
            const ifFalse = node.ifFalse
                ? (_c = visit(this)(node.ifFalse)) !== null && _c !== void 0 ? _c : undefined
                : undefined;
            if (ifFalse)
                (0, nodes_1.assertIsNode)(ifFalse, nodes_1.CONDITIONAL_VALUE_BRANCH_NODES);
            if (!ifTrue && !ifFalse)
                return null;
            return (0, nodes_1.conditionalValueNode)({ condition, value, ifTrue, ifFalse });
        };
    }
    if (castedNodeKeys.includes('pdaValueNode')) {
        visitor.visitPdaValue = function visitPdaValue(node) {
            const pda = visit(this)(node.pda);
            if (pda === null)
                return null;
            (0, nodes_1.assertIsNode)(pda, 'pdaLinkNode');
            const seeds = node.seeds
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('pdaSeedValueNode'));
            return (0, nodes_1.pdaValueNode)(pda, seeds);
        };
    }
    if (castedNodeKeys.includes('pdaSeedValueNode')) {
        visitor.visitPdaSeedValue = function visitPdaSeedValue(node) {
            const value = visit(this)(node.value);
            if (value === null)
                return null;
            (0, nodes_1.assertIsNode)(value, [
                ...nodes_1.VALUE_NODES,
                'accountValueNode',
                'argumentValueNode',
            ]);
            return (0, nodes_1.pdaSeedValueNode)(node.name, value);
        };
    }
    if (castedNodeKeys.includes('fixedSizeTypeNode')) {
        visitor.visitFixedSizeType = function visitFixedSizeType(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.fixedSizeTypeNode)(type, node.size);
        };
    }
    if (castedNodeKeys.includes('sizePrefixTypeNode')) {
        visitor.visitSizePrefixType = function visitSizePrefixType(node) {
            const prefix = visit(this)(node.prefix);
            if (prefix === null)
                return null;
            (0, nodes_1.assertIsNestedTypeNode)(prefix, 'numberTypeNode');
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.sizePrefixTypeNode)(type, prefix);
        };
    }
    if (castedNodeKeys.includes('preOffsetTypeNode')) {
        visitor.visitPreOffsetType = function visitPreOffsetType(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.preOffsetTypeNode)(type, node.offset, node.strategy);
        };
    }
    if (castedNodeKeys.includes('postOffsetTypeNode')) {
        visitor.visitPostOffsetType = function visitPostOffsetType(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.postOffsetTypeNode)(type, node.offset, node.strategy);
        };
    }
    if (castedNodeKeys.includes('sentinelTypeNode')) {
        visitor.visitSentinelType = function visitSentinelType(node) {
            const sentinel = visit(this)(node.sentinel);
            if (sentinel === null)
                return null;
            (0, nodes_1.assertIsNode)(sentinel, 'constantValueNode');
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.sentinelTypeNode)(type, sentinel);
        };
    }
    if (castedNodeKeys.includes('hiddenPrefixTypeNode')) {
        visitor.visitHiddenPrefixType = function visitHiddenPrefixType(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            const prefix = node.prefix
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('constantValueNode'));
            if (prefix.length === 0)
                return type;
            return (0, nodes_1.hiddenPrefixTypeNode)(type, prefix);
        };
    }
    if (castedNodeKeys.includes('hiddenSuffixTypeNode')) {
        visitor.visitHiddenSuffixType = function visitHiddenSuffixType(node) {
            const type = visit(this)(node.type);
            if (type === null)
                return null;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            const suffix = node.suffix
                .map(visit(this))
                .filter((0, nodes_1.removeNullAndAssertIsNodeFilter)('constantValueNode'));
            if (suffix.length === 0)
                return type;
            return (0, nodes_1.hiddenSuffixTypeNode)(type, suffix);
        };
    }
    if (castedNodeKeys.includes('constantDiscriminatorNode')) {
        visitor.visitConstantDiscriminator = function visitConstantDiscriminator(node) {
            const constant = visit(this)(node.constant);
            if (constant === null)
                return null;
            (0, nodes_1.assertIsNode)(constant, 'constantValueNode');
            return (0, nodes_1.constantDiscriminatorNode)(constant, node.offset);
        };
    }
    return visitor;
}
exports.identityVisitor = identityVisitor;
//# sourceMappingURL=identityVisitor.js.map