"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.updateInstructionsVisitor = void 0;
const nodes_1 = require("../nodes");
const shared_1 = require("../shared");
const bottomUpTransformerVisitor_1 = require("./bottomUpTransformerVisitor");
const fillDefaultPdaSeedValuesVisitor_1 = require("./fillDefaultPdaSeedValuesVisitor");
const recordLinkablesVisitor_1 = require("./recordLinkablesVisitor");
const visitor_1 = require("./visitor");
function updateInstructionsVisitor(map) {
    const linkables = new shared_1.LinkableDictionary();
    const transformers = Object.entries(map).map(([selector, updates]) => ({
        select: ['[instructionNode]', selector],
        transform: (node) => {
            (0, nodes_1.assertIsNode)(node, 'instructionNode');
            if ('delete' in updates) {
                return null;
            }
            const { accounts: accountUpdates, arguments: argumentUpdates, ...metadataUpdates } = updates;
            const { newArguments, newExtraArguments } = handleInstructionArguments(node, argumentUpdates !== null && argumentUpdates !== void 0 ? argumentUpdates : {});
            const newAccounts = node.accounts.map((account) => handleInstructionAccount(node, account, accountUpdates !== null && accountUpdates !== void 0 ? accountUpdates : {}, linkables));
            return (0, nodes_1.instructionNode)({
                ...node,
                ...metadataUpdates,
                accounts: newAccounts,
                arguments: newArguments,
                extraArguments: newExtraArguments.length > 0 ? newExtraArguments : undefined,
            });
        },
    }));
    return (0, shared_1.pipe)((0, bottomUpTransformerVisitor_1.bottomUpTransformerVisitor)(transformers), (v) => (0, recordLinkablesVisitor_1.recordLinkablesVisitor)(v, linkables));
}
exports.updateInstructionsVisitor = updateInstructionsVisitor;
function handleInstructionAccount(instruction, account, accountUpdates, linkables) {
    const accountUpdate = accountUpdates === null || accountUpdates === void 0 ? void 0 : accountUpdates[account.name];
    if (!accountUpdate)
        return account;
    const { defaultValue, ...acountWithoutDefault } = {
        ...account,
        ...accountUpdate,
    };
    if (!defaultValue) {
        return (0, nodes_1.instructionAccountNode)(acountWithoutDefault);
    }
    return (0, nodes_1.instructionAccountNode)({
        ...acountWithoutDefault,
        defaultValue: (0, visitor_1.visit)(defaultValue, (0, fillDefaultPdaSeedValuesVisitor_1.fillDefaultPdaSeedValuesVisitor)(instruction, linkables)),
    });
}
function handleInstructionArguments(instruction, argUpdates) {
    var _a;
    const usedArguments = new Set();
    const newArguments = instruction.arguments.map((node) => {
        var _a, _b, _c, _d, _e;
        const argUpdate = argUpdates[node.name];
        if (!argUpdate)
            return node;
        usedArguments.add(node.name);
        return (0, nodes_1.instructionArgumentNode)({
            ...node,
            type: (_a = argUpdate.type) !== null && _a !== void 0 ? _a : node.type,
            name: (_b = argUpdate.name) !== null && _b !== void 0 ? _b : node.name,
            docs: (_c = argUpdate.docs) !== null && _c !== void 0 ? _c : node.docs,
            defaultValue: (_d = argUpdate.defaultValue) !== null && _d !== void 0 ? _d : node.defaultValue,
            defaultValueStrategy: (_e = argUpdate.defaultValueStrategy) !== null && _e !== void 0 ? _e : node.defaultValueStrategy,
        });
    });
    const updatedExtraArguments = ((_a = instruction.extraArguments) !== null && _a !== void 0 ? _a : []).map((node) => {
        var _a, _b, _c, _d, _e;
        if (usedArguments.has(node.name))
            return node;
        const argUpdate = argUpdates[node.name];
        if (!argUpdate)
            return node;
        usedArguments.add(node.name);
        return (0, nodes_1.instructionArgumentNode)({
            ...node,
            type: (_a = argUpdate.type) !== null && _a !== void 0 ? _a : node.type,
            name: (_b = argUpdate.name) !== null && _b !== void 0 ? _b : node.name,
            docs: (_c = argUpdate.docs) !== null && _c !== void 0 ? _c : node.docs,
            defaultValue: (_d = argUpdate.defaultValue) !== null && _d !== void 0 ? _d : node.defaultValue,
            defaultValueStrategy: (_e = argUpdate.defaultValueStrategy) !== null && _e !== void 0 ? _e : node.defaultValueStrategy,
        });
    });
    const newExtraArguments = [
        ...updatedExtraArguments,
        ...Object.entries(argUpdates)
            .filter(([argName]) => !usedArguments.has(argName))
            .map(([argName, argUpdate]) => {
            var _a, _b, _c, _d;
            const { type } = argUpdate;
            (0, nodes_1.assertIsNode)(type, nodes_1.TYPE_NODES);
            return (0, nodes_1.instructionArgumentNode)({
                name: (_a = argUpdate.name) !== null && _a !== void 0 ? _a : argName,
                type,
                docs: (_b = argUpdate.docs) !== null && _b !== void 0 ? _b : [],
                defaultValue: (_c = argUpdate.defaultValue) !== null && _c !== void 0 ? _c : undefined,
                defaultValueStrategy: (_d = argUpdate.defaultValueStrategy) !== null && _d !== void 0 ? _d : undefined,
            });
        }),
    ];
    return { newArguments, newExtraArguments };
}
//# sourceMappingURL=updateInstructionsVisitor.js.map